from phosphorus.lib.parser import parse_args
from phosphorus.subcommands.build import BuildCommand
from phosphorus.subcommands.check import CheckCommand
from phosphorus.subcommands.install import InstallCommand
from phosphorus.subcommands.lock import LockCommand


def main() -> None:
    args = parse_args()
    match args.subcommand:
        case "build":
            BuildCommand(args).run()
        case "check":
            CheckCommand(args).run()
        case "install":
            InstallCommand(args).run()
        case "lock":
            LockCommand(args).run()



import sys
from argparse import ArgumentParser, BooleanOptionalAction, Namespace

from phosphorus.__version__ import __version__

sys.tracebacklimit = 0


def add_build_args(parser: ArgumentParser) -> None:
    parser.add_argument("--metadata", action=BooleanOptionalAction, default=False)
    parser.add_argument("--sdist", action=BooleanOptionalAction, default=True)
    parser.add_argument("--wheel", action=BooleanOptionalAction, default=True)


def parse_args() -> Namespace:
    parser = ArgumentParser(description="a dependency management tool")
    parser = ArgumentParser(description="git repository manager")
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"phosphorus {__version__}",
        help="print the version and exit",
    )

    parent_parser = ArgumentParser(add_help=False)
    parent_parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        dest="verbosity",
        help="increase the level of verbosity",
    )

    subparsers = parser.add_subparsers(dest="subcommand", required=True)

    build_parser = subparsers.add_parser(
        "build",
        parents=[parent_parser],
        help="build the wheel and the sdist distributions for the package",
    )
    build_parser.add_argument(
        "--sdist",
        action=BooleanOptionalAction,
        default=True,
        help="build the sdist distribution",
    )
    build_parser.add_argument(
        "--wheel",
        action=BooleanOptionalAction,
        default=True,
        help="build the wheel distribution",
    )

    subparsers.add_parser(
        "check", parents=[parent_parser], help="check that the lock file is up to date"
    )

    install_parser = subparsers.add_parser(
        "install", parents=[parent_parser], help="install the project dependencies"
    )
    install_parser.add_argument(
        "--sync",
        action=BooleanOptionalAction,
        default=False,
        help="sync the dependencies, by removing the ones that are not in the lock file",
    )
    dependency_groups = install_parser.add_mutually_exclusive_group()
    dependency_groups.add_argument(
        "-e",
        "--exclude",
        action="append",
        default=[],
        help="exclude a group of dependencies",
    )
    dependency_groups.add_argument(
        "-g",
        "--groups",
        action="append",
        default=[],
        help="install a group of dependencies",
    )

    lock_parser = subparsers.add_parser(
        "lock", parents=[parent_parser], help="lock the project dependencies"
    )
    lock_parser.add_argument(
        "-f", "--force", action="store_true", help="force recreating the lock"
    )
    lock_parser.add_argument(
        "--enforce-pep440",
        action=BooleanOptionalAction,
        default=True,
        help="enforce PEP 440",
    )
    lock_parser.add_argument(
        "--allow-pre-releases",
        action=BooleanOptionalAction,
        default=False,
        help="allow pre-releases",
    )
    lock_parser.add_argument(
        "--allow-dev-releases",
        action=BooleanOptionalAction,
        default=False,
        help="allow dev releases",
    )

    args = parser.parse_args()
    if args.verbosity > 0:
        sys.tracebacklimit = 1000

    return args



import sys
from tomllib import TOMLDecodeError

from phosphorus.subcommands.base import BaseCommand


class CheckCommand(BaseCommand):
    __slots__ = ()

    def run(self) -> None:
        try:
            self.check()
        except RuntimeError as exc:
            print(f"❌ Lockfile check failed: {exc}", file=sys.stderr)
            sys.exit(1)
        else:
            print("✅ Success: Lockfile is valid and up-to-date with pyproject.toml")

    def check(self) -> None:
        try:
            p_hash = self.get_current_hash()
        except FileNotFoundError as exc:
            msg = "No lockfile found"
            raise RuntimeError(msg) from exc
        except (KeyError, TOMLDecodeError) as exc:
            msg = "Lockfile is invalid"
            raise RuntimeError(msg) from exc

        if p_hash != self.meta.p_hash:
            msg = "Hashes don't match"
            raise RuntimeError(msg)



from __future__ import annotations

import json
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any
from urllib.error import HTTPError
from urllib.request import Request, urlopen

from phosphorus.lib.constants import pypi_cache
from phosphorus.lib.utils import canonicalise_name
from phosphorus.lib.versions import Version, VersionClause

if TYPE_CHECKING:
    from collections.abc import Iterable

    from phosphorus.lib.requirements import Requirement


@dataclass(frozen=True, slots=True, order=True)
class Package:
    name: str
    distribution_name: str = field(repr=False, compare=False)

    def __init__(self, name: str) -> None:
        name = canonicalise_name(name)
        object.__setattr__(self, "name", name)
        object.__setattr__(self, "distribution_name", name.replace("-", "_"))

    def __str__(self) -> str:
        return self.name

    def get_versions(
        self,
        *,
        enforce_pep440: bool = False,
        allow_pre_releases: bool = True,
        allow_dev_releases: bool = True,
        clauses: Iterable[VersionClause] = (),
        last_check: float = float("inf"),
    ) -> list[Version]:
        dependencies = self._get_version_info(last_check=last_check)
        versions: list[Version] = []
        for release in dependencies["releases"]:
            version = Version.from_string(release)
            if version.is_pre_release and not allow_pre_releases:
                continue
            if version.is_dev_release and not allow_dev_releases:
                continue
            if not version.pep_440_compliant and enforce_pep440:
                continue
            for clause in clauses:
                if not clause.match(version):
                    break
            else:
                versions.append(version)
        return sorted(versions)

    def get_version_info(
        self, version: Version, *, last_check: float = float("inf")
    ) -> ResolvedPackage:
        from phosphorus.lib.requirements import Requirement

        info = self._get_version_info(version.canonical_form, last_check=last_check)
        return ResolvedPackage(
            package=self,
            version=version,
            yanked=info["yanked"],
            requires_python=[
                VersionClause.from_string(version.strip())
                for version in (info["requires_python"] or "*").split(",")
            ],
            requires_dist=[
                Requirement.from_string(requirement)
                for requirement in info["requires_dist"] or []
            ],
        )

    def _get_version_info(
        self, version: str = "", *, last_check: float = float("inf")
    ) -> dict[str, Any]:
        if version:
            version = Version.from_string(version).canonical_form
            filename = f"{version}.json"
        else:
            filename = "latest.json"
        local_cache_dir = pypi_cache.joinpath(self.name)
        local_cache_dir.mkdir(parents=True, exist_ok=True)
        local_cache = local_cache_dir.joinpath(filename)
        etag = ""
        if local_cache.exists():
            with local_cache.open() as file:
                local_info: dict[str, Any] = json.load(file)
                etag = local_info["ETag"]
            modified = local_cache.stat().st_mtime
            if modified >= last_check:
                return local_info

        pypi_info = self._get_info_from_pypi(etag=etag, version=version)
        if pypi_info["use_cache"]:
            local_cache.touch()
            if not version:
                version = Version.from_string(local_info["version"]).canonical_form
                local_cache_dir.joinpath(f"{version}.json").touch()
            return local_info

        del pypi_info["use_cache"]
        with local_cache.open("w") as file:
            json.dump(pypi_info, file, indent=4)
        if not version:
            version = Version.from_string(pypi_info["version"]).canonical_form
            local_cache = local_cache_dir.joinpath(f"{version}.json")
            with local_cache.open("w") as file:
                json.dump(pypi_info, file, indent=4)
        return pypi_info

    def _get_info_from_pypi(self, etag: str, version: str) -> dict[str, Any]:
        package_slug = f"{self.name}/{version}" if version else self.name
        url = f"https://pypi.org/pypi/{package_slug}/json"
        headers = {}
        if etag:
            headers["If-None-Match"] = etag
        request = Request(url, headers=headers)  # noqa: S310
        try:
            response = urlopen(request, timeout=60)  # noqa: S310
        except TimeoutError as exc:
            msg = "pypi connection timed out"
            raise RuntimeError(msg) from exc
        except HTTPError as exc:
            if exc.status == 304:
                return {"use_cache": True}
            raise
        info = json.loads(response.read())
        out = {
            "use_cache": False,
            "ETag": response.headers["ETag"],
            "requires_dist": info["info"]["requires_dist"],
            "requires_python": info["info"]["requires_python"],
            "yanked": info["info"]["yanked"],
            "version": info["info"]["version"],
        }
        if not version:
            out["releases"] = list(info["releases"].keys())
        return out


@dataclass(frozen=True, slots=True, order=True)
class ResolvedPackage:
    package: Package
    version: Version
    yanked: bool
    requires_python: list[VersionClause]
    requires_dist: list[Requirement]



import tarfile
from collections.abc import Iterator
from pathlib import Path

from phosphorus.construction.base import Builder
from phosphorus.lib.constants import pyproject_base_name
from phosphorus.lib.licenses import get_license_files
from phosphorus.lib.zipped_file import ArchiveFile


class SdistBuilder(Builder):
    __slots__ = ()

    @property
    def filename(self) -> str:
        return f"{self.base_name}.tar.gz"

    def collect_files(self, temp_dir: Path) -> list[ArchiveFile]:
        output: list[ArchiveFile] = []
        base_dir = self.meta.base_dir
        for package in self.meta.package_paths:
            output.extend(
                ArchiveFile.from_file(source=file, target=file.relative_to(base_dir))
                for file in package.path.glob("**/*")
                if file.is_file()
            )
        output.extend(
            ArchiveFile.from_file(source=source, target=target)
            for source, target in self.non_package_files(temp_dir)
        )
        return output

    def write_files(
        self, files: list[ArchiveFile], package: Path, _temp_dir: Path
    ) -> None:
        with tarfile.open(package, "w:gz") as tar:
            for archive_file in sorted(files):
                tar.addfile(
                    archive_file.tar_info, archive_file.absolute_path.open("rb")
                )

    def non_package_files(self, temp_dir: Path) -> Iterator[tuple[Path, Path]]:
        base_dir = self.meta.base_dir

        yield base_dir.joinpath(pyproject_base_name), Path(pyproject_base_name)

        for license_file in get_license_files(base_dir):
            yield license_file, license_file.relative_to(base_dir)

        if self.meta.readme.read_text():
            yield self.meta.readme, self.meta.readme.relative_to(base_dir)

        pkg_info = temp_dir.joinpath("PKG-INFO")
        with pkg_info.open("w") as file:
            file.writelines(f"{line}\n" for line in self.get_metadata_content())

        yield pkg_info, pkg_info.relative_to(temp_dir)



from __future__ import annotations

import sys
import tomllib
from argparse import Namespace
from subprocess import run
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING
from warnings import warn

from phosphorus.lib.constants import hash_prefix
from phosphorus.lib.exceptions import ThirdPartyError
from phosphorus.lib.markers import Marker
from phosphorus.lib.packages import Package
from phosphorus.lib.requirements import ResolvedRequirement
from phosphorus.lib.resolver import LockEntry
from phosphorus.lib.term import SGRParams, SGRString
from phosphorus.lib.versions import Version
from phosphorus.subcommands.base import BaseCommand
from phosphorus.subcommands.lock import LockCommand

if TYPE_CHECKING:
    from phosphorus.lib.types import PackageDiff


class InstallCommand(BaseCommand):
    __slots__ = ("groups", "sync")

    def __init__(self, args: Namespace, /) -> None:
        super().__init__(args)
        self.sync = args.sync
        self.groups = self.get_groups(set(args.groups), set(args.exclude))

    def run(self) -> None:
        try:
            self.install()
        except RuntimeError as exc:
            print(f"❌ Installation failed: {exc}", file=sys.stderr)
            sys.exit(1)
        else:
            print("✅ Success: Installation finished successfully!")

    def install(self) -> None:
        if not self.groups:
            msg = "No remaining valid groups"
            raise RuntimeError(msg)

        if not self.meta.lockfile.exists():
            print("🔒 Lockfile not found, creating one...")
            LockCommand(
                Namespace(
                    enforce_pep440=True,
                    allow_pre_releases=False,
                    allow_dev_releases=False,
                    force=False,
                )
            ).run()

        package_diff = self.get_package_diff()

        print("📦 Installing packages...")
        for old_package, lock_entry in package_diff["update"]:
            self.install_package(old_package, lock_entry)
        if self.sync:
            print("🧹 Removing packages...")
            for old_package in package_diff["remove"]:
                self.remove_package(old_package)
        self.install_self()

    def get_groups(self, included: set[str], excluded: set[str]) -> set[str]:
        all_groups = {group.group for group in self.meta.requirement_groups}
        if included:
            if not included.issubset(all_groups):
                included &= all_groups
                msg = f"Unknown group(s) {included}, ignoring..."
                warn(msg, RuntimeWarning, stacklevel=3)
            return included
        if not excluded.issubset(all_groups):
            msg = f"Unknown group(s) {excluded - all_groups}, ignoring..."
            warn(msg, RuntimeWarning, stacklevel=3)
        return all_groups - excluded

    def get_package_diff(self) -> PackageDiff:
        print("🔎 Looking for packages to install/remove/upgrade...")

        output = run(  # noqa: PLW1510, S603
            [  # noqa: S607
                "uv",
                "pip",
                "freeze",
                "--exclude-editable",
            ],
            capture_output=True,
        )
        existing_venv = {
            resolved_requirement.package: resolved_requirement
            for requirement in output.stdout.decode().splitlines()
            if (resolved_requirement := ResolvedRequirement.from_string(requirement))
        }

        with self.meta.lockfile.open("rb") as lockfile:
            lock = tomllib.load(lockfile)

        target_venv = {}
        for lock_entry in lock["packages"]:
            marker = Marker.from_string(lock_entry.get("marker", ""))
            if not marker.evaluate():
                continue
            if not self.groups & set(lock_entry["groups"]):
                continue
            package = Package(lock_entry["name"])
            entry = LockEntry(
                requirement=ResolvedRequirement(
                    package=package,
                    version=Version.from_string(lock_entry["version"]),
                    marker=marker,
                ),
                groups=tuple(lock_entry["groups"]),
                hashes=tuple(lock_entry["hashes"]),
            )
            target_venv[package] = entry

        update: set[tuple[ResolvedRequirement | None, LockEntry]] = set()
        remove = set()

        for package in existing_venv | target_venv:
            if package not in target_venv:
                remove.add(existing_venv[package])
            elif package not in existing_venv:
                update.add((None, target_venv[package]))
            else:
                existing = existing_venv[package]
                target = target_venv[package]
                if existing.version != target.requirement.version:
                    update.add((existing, target))

        return {"update": update, "remove": remove}

    def install_package(
        self, old_package: ResolvedRequirement | None, lock_entry: LockEntry
    ) -> None:
        package = lock_entry.requirement.package
        old_version = old_package.version if old_package else None
        new_version = lock_entry.requirement.version
        hashes = " ".join(
            f"{hash_prefix}{hashed_file}" for hashed_file in lock_entry.hashes
        )
        if not old_version:
            action = "Installing"
            version = str(new_version)
        elif old_version < new_version:
            action = "Upgrading"
            version = f"{old_version} → {new_version}"
        else:
            action = "Downgrading"
            version = f"{old_version} → {new_version}"
        print(
            f"⏳ {action} {SGRString(str(package), params=[SGRParams.CYAN])} "
            f"({SGRString(version, params=[SGRParams.BLUE_BRIGHT])})...",
            end=" ",
        )
        with NamedTemporaryFile("w", delete=False) as tmp:
            line = f"{package}=={new_version} {hashes}"
            tmp.write(f"{line}\n")
            tmp.seek(0)
            output = run(  # noqa: PLW1510, S603
                [  # noqa: S607
                    "uv",
                    "pip",
                    "install",
                    "--no-deps",
                    "--require-hashes",
                    "-r",
                    tmp.name,
                ],
                capture_output=True,
            )
            if output.returncode == 0:
                print("🗸")
            else:
                print()
                command = "uv pip install"
                raise ThirdPartyError(command)

    def remove_package(self, old_package: ResolvedRequirement) -> None:
        package = old_package.package
        version = old_package.version
        print(
            f"⏳ Removing {SGRString(str(package), params=[SGRParams.CYAN])} "
            f"({SGRString(str(version), params=[SGRParams.BLUE_BRIGHT])})...",
            end=" ",
        )
        output = run(  # noqa: PLW1510, S603
            [  # noqa: S607
                "uv",
                "pip",
                "uninstall",
                package.name,
            ],
            capture_output=True,
        )
        if output.returncode == 0:
            print("🗸")
        else:
            print()
            command = "uv pip uninstall"
            raise ThirdPartyError(command)

    def install_self(self) -> None:
        print(
            f"⏳ Installing {SGRString(str(self.meta.package), params=[SGRParams.CYAN])} "
            f"({SGRString(str(self.meta.version), params=[SGRParams.BLUE_BRIGHT])})...",
            end=" ",
        )
        output = run(  # noqa: PLW1510, S603
            [  # noqa: S607
                "uv",
                "pip",
                "install",
                "--no-deps",
                "--editable",
                ".",
            ],
            capture_output=True,
        )
        if output.returncode == 0:
            print("🗸")
        else:
            print()
            command = "uv pip install"
            raise ThirdPartyError(command)
